import datetime
import sys
import time
import zipfile
from pathlib import Path

import httpx
import typer
import nbformat
from rich import print
from rich.console import Console
from rich.panel import Panel

from taas_client import __version__

console = Console()
app = typer.Typer()

BASE_URL = "http://localhost:5000"
PROJECT_NB_NAME = "project.ipynb"
PROJECT_TOPIC_COUNT = 2
PROJECT_LIMIT_SOFT_MB = 5
PROJECT_LIMIT_HARD_MB = 50


class CheckError(Exception):
    pass


def version_callback(value: bool) -> None:
    if not value:
        return

    print(f"taas-client: {__version__}")

    # FIXME how could we use the given URL argument here too?
    r = httpx.get(f"{BASE_URL}/version")
    r.raise_for_status()
    server_version = r.json()["version"]
    print(f"taas server: {server_version}")

    raise typer.Exit()


def check_notebook_file(path: Path, force: bool) -> None:
    if not path.is_relative_to(Path.home() / "work" / "AutPy"):
        console.print(
            f"[red]Refusing to submit file from unexpected location[/]: {path}\n"
            "Make sure you're not accidentally working in the 'Originale' folder!"
        )
        if not force:
            raise CheckError

    modified = round(time.time() - path.stat().st_mtime)
    sec = "second" if modified == 1 else "seconds"
    suffix = ", [red]make sure you saved![/]" if modified >= 10 else ""
    console.print(f"Last change: [bold]{modified}[/bold] {sec} ago{suffix}")


def find_nb_tag(nb: nbformat.NotebookNode, tag: str) -> nbformat.NotebookNode:
    for cell in nb.cells:
        assert isinstance(cell, nbformat.NotebookNode)
        if tag in cell.metadata.get("tags", []):
            return cell
    console.print(f"Missing {tag} cell in notebook")
    raise CheckError


def check_project_file(path: Path, force: bool) -> None:
    size = path.stat().st_size
    soft_limit = PROJECT_LIMIT_SOFT_MB * 1024 * 1024
    hard_limit = PROJECT_LIMIT_HARD_MB * 1024 * 1024
    assert hard_limit > soft_limit

    if size > soft_limit:
        console.print(
            ":x: [red]Your project is over the size limit of "
            f"{PROJECT_LIMIT_SOFT_MB} MB.[/red]\n\n"
            "This doesn't sound like much, but things add up with >100 students and "
            "multiple submissions each. For source code, this should be plenty.\n\n"
            "[b]Please make sure[/b] your submission does not accidentally contain "
            "things like dog pictures/videos or your virtual environment.\n"
        )

        if size > hard_limit:
            console.print(
                ":bomb: Since your file is also [b]above the hard limit[/] of "
                f"{PROJECT_LIMIT_HARD_MB} MB, this error can not be overridden.\n\n"
                "[b]Please reduce the file size or contact your instructor.[/]"
            )
            raise CheckError

        if not force:
            console.print(
                f"[grey]You can use [b]!submit --project --force {path.name}[/b] "
                "to override this warning and submit anyways.\n"
                "Please only do so if absolutely necessary (e.g. shortly before the "
                "deadline).[/]\n\n"
                "Otherwise, please [b]reduce the file size[/], "
                "or [b]contact your instructor[/] if you need assistance."
            )
            raise CheckError
        else:
            console.print("Continuing submission since --force has been given.\n")

    if not zipfile.is_zipfile(path):
        console.print(f"[red]{path.name} is not a valid .zip file.[/]")
        raise CheckError

    with open(PROJECT_NB_NAME, "r") as f:
        nb = nbformat.read(f, as_version=4)

    topics_cell = find_nb_tag(nb, "selected-topics")
    cross_count = topics_cell.source.count("- [x]")
    if cross_count != PROJECT_TOPIC_COUNT:
        console.print(
            f":warning: Found {cross_count} selected topics in the "
            "'Selected additional topics' cell, but expected exactly "
            f"{PROJECT_TOPIC_COUNT}.\n"
            "[b]Please make sure to fill out the list properly and submit again.[/b]"
        )


@app.command()
def upload_file(
    notebook: Path = typer.Argument(
        ...,
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=False,
        readable=True,
        resolve_path=True,
    ),
    url: str = typer.Argument(f"{BASE_URL}/testbook", envvar="TAAS_URL"),
    project: bool = False,
    force: bool = False,
    _: bool = typer.Option(False, "--version", callback=version_callback),
) -> None:
    if project:
        # dirty hack because url is fixed for testbook at the moment.
        # After the module it needs to be changed to a base url
        url = f"{url[:-8]}project"
    timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    try:
        if project:
            check_project_file(path=notebook, force=force)
        else:
            check_notebook_file(path=notebook, force=force)
    except CheckError:
        print(f":boom: [bold]Submission failed![/] [white]({timestamp})[/]")
        raise typer.Exit(code=1)

    files = {"file": open(notebook, "rb")}
    if project:
        files["notebook"] = open(PROJECT_NB_NAME, "rb")

    try:
        text = "Uploading" if project else "Testing"
        with console.status(f"[bold green]{text}..."):
            r = httpx.post(url, files=files, timeout=600.0)
        r.raise_for_status()

        result = r.json()
        test_result = result["tests"]
        for status, style in [
            ("error", "bold red"),
            ("failed", "red"),
            ("passed", "green"),
            ("skipped", ""),
            ("xfailed", "red"),
            ("xpassed", ""),
        ]:
            if test_result[status]:
                lines = [
                    f" -  test_{test['nodeid']} {test['outcome']} at {test['when']}"
                    if test["when"] != "call"
                    else f" -  test_{test['nodeid']} {test['outcome']}"
                    for test in test_result[status]
                ]
                console.print(
                    Panel(
                        "\n".join(lines),
                        title=status.capitalize(),
                        subtitle=f"{len(test_result[status])}",
                        style=style,
                    )
                )
    except httpx.HTTPError as exc:
        print(f"[red bold]{exc}[/]", file=sys.stderr)
        try:
            data = exc.response.json()  # type: ignore[attr-defined]
            print(f"[red]Error: {data.get('error')}[/]", file=sys.stderr)
            print(f"[red]Message: {data.get('msg')}[/]", file=sys.stderr)
            print(f":boom: [bold]Submission failed![/] [white]({timestamp})[/]")
        except Exception:  # FIXME
            pass
        raise typer.Exit(code=1)

    commit = result.get("commit", "?")
    print(
        ":mailbox: :heavy_check_mark: [bold]Submission successful![/] "
        f"[white]({timestamp}, {commit})[/]"
    )


if __name__ == "__main__":
    app()  # pragma: no cover
